SHELL=/bin/sh

.SUFFIXES:
.SUFFIXES: .cpp .o

CC=g++
AR=ar cr

CFLAGS=-Wall -O2
CFLAGSD=-Wall -g -DDEBUG
#CFLAGSD=-Wall -g -DDEBUG -DCVM_FLOAT

# replace flags located above with the following ones if you want to use STLport
#CFLAGS=-Wall -O2 -I/usr/include/stlport
#CFLAGSD=-Wall -g -DDEBUG -I/usr/include/stlport

SOCFLAGS=$(CFLAGS) -D_REENTRANT -DCVM_EXPORTS -fPIC
SOCFLAGSD=$(CFLAGSD) -D_REENTRANT -DCVM_EXPORTS -fPIC
SONAME=libcvm.so.5.2
SODNAME=libcvm_debug.so.5.2
SO_MKL_NAME=libcvm_mkl.so.5.2
SOD_MKL_NAME=libcvm_mkl_debug.so.5.2

# it is assumed that you already have BLAS & LAPACK packages installed
LAPACK_BLAS_OPT=-llapack -lblas

# alternatively you can purchase or evaluate Intel MKL 6.0 library:
INTEL_MKL_PATH=/opt/intel/mkl72/lib/32
INTEL_MKL_OPT=-lmkl_lapack -lmkl_ia32 -lguide


all : libcvm.a libcvm_debug.a $(SONAME) $(SODNAME) $(SO_MKL_NAME) $(SOD_MKL_NAME)

release : libcvm.a

debug : libcvm_debug.a

so : $(SONAME) $(SODNAME)

so_release : $(SONAME)

so_debug : $(SODNAME)

so_mkl : $(SO_MKL_NAME) $(SOD_MKL_NAME)

so_mkl_release : $(SO_MKL_NAME)

so_mkl_debug : $(SOD_MKL_NAME)


OBJDIR=./obj
DOBJDIR=./dobj
SOOBJDIR=./soobj
SODOBJDIR=./sodobj
SO_MKL_OBJDIR=./so_mkl_obj
SOD_MKL_OBJDIR=./sod_mkl_obj
OUTDIR=../lib


OBJS= \
    $(OBJDIR)/globals.o \
    $(OBJDIR)/utils.o \
    $(OBJDIR)/arrays.o \
    $(OBJDIR)/cmatrix.o \
    $(OBJDIR)/cvector.o \
    $(OBJDIR)/rmatrix.o \
    $(OBJDIR)/rvector.o \
    $(OBJDIR)/scmatrix.o \
    $(OBJDIR)/srmatrix.o

DOBJS= \
    $(DOBJDIR)/globals.o \
    $(DOBJDIR)/utils.o \
    $(DOBJDIR)/arrays.o \
    $(DOBJDIR)/cmatrix.o \
    $(DOBJDIR)/cvector.o \
    $(DOBJDIR)/rmatrix.o \
    $(DOBJDIR)/rvector.o \
    $(DOBJDIR)/scmatrix.o \
    $(DOBJDIR)/srmatrix.o

SOOBJS= \
    $(SOOBJDIR)/globals.o \
    $(SOOBJDIR)/utils.o \
    $(SOOBJDIR)/arrays.o \
    $(SOOBJDIR)/cmatrix.o \
    $(SOOBJDIR)/cvector.o \
    $(SOOBJDIR)/rmatrix.o \
    $(SOOBJDIR)/rvector.o \
    $(SOOBJDIR)/scmatrix.o \
    $(SOOBJDIR)/srmatrix.o

SODOBJS= \
    $(SODOBJDIR)/globals.o \
    $(SODOBJDIR)/utils.o \
    $(SODOBJDIR)/arrays.o \
    $(SODOBJDIR)/cmatrix.o \
    $(SODOBJDIR)/cvector.o \
    $(SODOBJDIR)/rmatrix.o \
    $(SODOBJDIR)/rvector.o \
    $(SODOBJDIR)/scmatrix.o \
    $(SODOBJDIR)/srmatrix.o

SO_MKL_OBJS= \
    $(SO_MKL_OBJDIR)/globals.o \
    $(SO_MKL_OBJDIR)/utils.o \
    $(SO_MKL_OBJDIR)/arrays.o \
    $(SO_MKL_OBJDIR)/cmatrix.o \
    $(SO_MKL_OBJDIR)/cvector.o \
    $(SO_MKL_OBJDIR)/rmatrix.o \
    $(SO_MKL_OBJDIR)/rvector.o \
    $(SO_MKL_OBJDIR)/scmatrix.o \
    $(SO_MKL_OBJDIR)/srmatrix.o

SOD_MKL_OBJS= \
    $(SOD_MKL_OBJDIR)/globals.o \
    $(SOD_MKL_OBJDIR)/utils.o \
    $(SOD_MKL_OBJDIR)/arrays.o \
    $(SOD_MKL_OBJDIR)/cmatrix.o \
    $(SOD_MKL_OBJDIR)/cvector.o \
    $(SOD_MKL_OBJDIR)/rmatrix.o \
    $(SOD_MKL_OBJDIR)/rvector.o \
    $(SOD_MKL_OBJDIR)/scmatrix.o \
    $(SOD_MKL_OBJDIR)/srmatrix.o


libcvm.a : $(OBJS)
	$(AR) $@ $(OBJS)
	ranlib $@
	mv $@ $(OUTDIR)

libcvm_debug.a : $(DOBJS)
	$(AR) $@ $(DOBJS)
	ranlib $@
	mv $@ $(OUTDIR)


$(SONAME) : $(SOOBJS)
	$(CC) -L../lib -shared -Wl,-soname,$(SONAME) -o $@ $(SOOBJS) -lftn $(LAPACK_BLAS_OPT) -lg2c -lpthread
	mv $@ $(OUTDIR)
	ln -sf $(OUTDIR)/$@ $(OUTDIR)/libcvm.so

$(SODNAME) : $(SODOBJS)
	$(CC) -L../lib -shared -Wl,-soname,$(SODNAME) -o $@ $(SODOBJS) -lftn $(LAPACK_BLAS_OPT) -lg2c -lpthread
	mv $@ $(OUTDIR)
	ln -sf $(OUTDIR)/$@ $(OUTDIR)/libcvm_debug.so

$(SO_MKL_NAME) : $(SO_MKL_OBJS)
	$(CC) -L../lib -shared -Wl,-soname,$(SO_MKL_NAME) -o $@ $(SO_MKL_OBJS) -L$(INTEL_MKL_PATH) $(INTEL_MKL_OPT) -lftn -lg2c -lpthread
	mv $@ $(OUTDIR)
	ln -sf $(OUTDIR)/$@ $(OUTDIR)/libcvm_mkl.so

$(SOD_MKL_NAME) : $(SOD_MKL_OBJS)
	$(CC) -L../lib -shared -Wl,-soname,$(SOD_MKL_NAME) -o $@ $(SOD_MKL_OBJS) -L$(INTEL_MKL_PATH) $(INTEL_MKL_OPT) -lftn -lg2c -lpthread
	mv $@ $(OUTDIR)
	ln -sf $(OUTDIR)/$@ $(OUTDIR)/libcvm_mkl_debug.so


$(OBJDIR)/%.o: %.cpp cvm.h
	@if [ ! -d $(OBJDIR) ]; then mkdir $(OBJDIR); fi
	$(CC) $(CFLAGS) -c -o $@ $<

$(DOBJDIR)/%.o: %.cpp cvm.h
	@if [ ! -d $(DOBJDIR) ]; then mkdir $(DOBJDIR); fi
	$(CC) $(CFLAGSD) -c -o $@ $<

$(SOOBJDIR)/%.o: %.cpp cvm.h
	@if [ ! -d $(SOOBJDIR) ]; then mkdir $(SOOBJDIR); fi
	$(CC) $(SOCFLAGS) -c -o $@ $<

$(SODOBJDIR)/%.o: %.cpp cvm.h
	@if [ ! -d $(SODOBJDIR) ]; then mkdir $(SODOBJDIR); fi
	$(CC) $(SOCFLAGSD) -c -o $@ $<

$(SO_MKL_OBJDIR)/%.o: %.cpp cvm.h
	@if [ ! -d $(SO_MKL_OBJDIR) ]; then mkdir $(SO_MKL_OBJDIR); fi
	$(CC) $(SOCFLAGS) -c -o $@ $<

$(SOD_MKL_OBJDIR)/%.o: %.cpp cvm.h
	@if [ ! -d $(SOD_MKL_OBJDIR) ]; then mkdir $(SOD_MKL_OBJDIR); fi
	$(CC) $(SOCFLAGSD) -c -o $@ $<


clean:
	rm -rf $(OBJS) $(DOBJS) $(SOOBJS) $(SODOBJS) $(SO_MKL_OBJS) $(SOD_MKL_OBJS)

cleanall:
	rm -rf $(OBJS) $(DOBJS) $(SOOBJS) $(SODOBJS) $(SO_MKL_OBJS) $(SOD_MKL_OBJS) $(OUTDIR)/libcvm.a $(OUTDIR)/libcvm_debug.a $(OUTDIR)/libcvm.so $(OUTDIR)/libcvm_debug.so $(OUTDIR)/libcvm_mkl.a $(OUTDIR)/libcvm_mkl_debug.a $(OUTDIR)/libcvm_mkl.so $(OUTDIR)/libcvm_mkl_debug.so
